#include "intelhex.h"

// Get the record type from an incoming line
uint8_t ihex_get_record_type (const char *in) {
    
    uint8_t rtype;
    
    // Valid line?
    if (*(in++) != ':') return IHEX_RTYPE_INVALID;
    
    // Skip len and address fields
    in += 6;
    
    // Get record type
    rtype = ihex_int(in);
    
    if (rtype > IHEX_RTYPE_STARTADDR32)
        return IHEX_RTYPE_INVALID;
    
    return rtype;
}


// Encodes a record line
void ihex_encode_raw (const char *data, char *out, uint8_t len, uint16_t offset, uint8_t rtype) {
    
    uint8_t cksum;
    
    // Start line
    *(out++) = ':';
    
    // Encode data length
    ihex_hex(out, len);
    out += 2;
    
    // Encode address offset
    ihex_hex(out, ((offset & 0xFF00) >> 8));
    out += 2;
    
    ihex_hex(out, (uint8_t)(offset & 0x00FF));
    out += 2;
    
    // Encode record type
    ihex_hex(out, rtype);
    out += 2;
    
    // Seed checksum with record preamble
    cksum = len + rtype
          + (uint8_t)((offset & 0xFF00) >> 8) + (uint8_t)(offset & 0x00FF);
    
    // Encode data
    while (len) {
        
        // Update checksum
        cksum += *((uint8_t*)data);
        
        // Encode data
        ihex_hex(out, (uint8_t)*(data++));
        
        out += 2;
        len--;
    }
    
    // Encode checksum
    cksum = (~cksum) + 1;
    
    ihex_hex(out, cksum);
    out += 2;
    
    // Terminate buffer
    *out = 0x00;
}


// Encodes a valid End-of-File record
void ihex_encode_eof (char *out) {
    
    // Let the raw line encoder do the heavy lifting
    ihex_encode_raw(NULL, out, 0, 0x0000, IHEX_RTYPE_EOF);
}


// Encodes an Extended Segment Address or Extended Linear Address record
void ihex_encode_ext_addr (char *out, uint8_t rtype, uint16_t addr) {
    
    char data[2];
    
    // Handle invalid record types
    if ( (rtype != IHEX_RTYPE_EXTADDR16) && (rtype != IHEX_RTYPE_EXTADDR32) ) {
        out = 0x0;
        return;
    }
    
    // Convert address field into binary data
    data[0] = (char)((addr & 0xFF00) >> 8);
    data[1] = (char)(addr & 0x00FF);
    
    // Encode as a raw line
    ihex_encode_raw(data, out, 2, 0x0000, rtype);
}


// Returns 16-bit Ext. Segment Addr or high word of 32-bit Ext. Linear Addr
uint16_t ihex_decode_ext_addr (const char *in) {
    
    uint8_t  rtype;
    uint8_t  cksum;
    uint16_t addr;
    uint16_t offset;
    
    // Valid line?
    if (*(in++) != ':') return 0xFFFF;
    
    // Len field should be == 2
    if (ihex_int(in) != 2) return 0xFFFF;
    in += 2;
    
    // Save address offset for checksum
    offset = (ihex_int(in) << 8) | ihex_int(in + 2);
    in += 4;
    
    // Get record type
    rtype = ihex_int(in);
    in += 2;
    
    // Handle invalid record types
    if ( (rtype != IHEX_RTYPE_EXTADDR16) && (rtype != IHEX_RTYPE_EXTADDR32) )
        return 0xFFFF;
    
    // Get address data (big endian)
    addr = (ihex_int(in) << 8) | ihex_int(in + 2);
    in += 4;
    
    // Verify checksum
    cksum = 2 + rtype
          + (uint8_t)((offset & 0xFF00) >> 8) + (uint8_t)(offset & 0x00FF)
          + (uint8_t)((addr & 0xFF00) >> 8) + (uint8_t)(addr & 0x00FF);
    
    cksum = (~cksum) + 1;
    
    if (cksum != ihex_int(in)) return 0xFFFF;
    
    return addr;
}


// Encodes an Start Segment Address or Start Linear Address record
void ihex_encode_start_addr (char *out, uint8_t rtype, uint32_t addr) {
    
    char data[4];
    
    // Handle invalid record types
    if ( (rtype != IHEX_RTYPE_STARTADDR16) &&
         (rtype != IHEX_RTYPE_STARTADDR32) )
    {
        out = 0x0;
        return;
    }
    
    // Convert address field into binary data
    data[0] = (char)((addr & 0xFF000000) >> 24);
    data[1] = (char)((addr & 0x00FF0000) >> 16);
    data[2] = (char)((addr & 0x0000FF00) >> 8);
    data[3] = (char)(addr & 0x000000FF);
    
    // Encode as a raw line
    ihex_encode_raw(data, out, 4, 0x0000, rtype);
}


// Returns 32-bit Start Segment (CS:IP) or Linear (EIP) Address
uint32_t ihex_decode_start_addr (const char *in) {
    
    uint8_t  rtype;
    uint8_t  cksum;
    uint16_t offset;
    uint32_t addr;
    
    // Valid line?
    if (*(in++) != ':') return 0xFFFFFFFF;
    
    // Len field should be == 4
    if (ihex_int(in) != 4) return 0xFFFFFFFF;
    in += 2;
    
    // Save address offset for checksum
    offset = (ihex_int(in) << 8) | ihex_int(in + 2);
    in += 4;
    
    // Get record type
    rtype = ihex_int(in);
    in += 2;
    
    // Must be Extended Segment Address or Extended Linear Address record
    if ( (rtype != IHEX_RTYPE_STARTADDR16) &&
         (rtype != IHEX_RTYPE_STARTADDR32) ) return 0xFFFFFFFF;
    
    // Get address data (big endian)
    addr = (ihex_int(in + 0) << 24) | (ihex_int(in + 2) << 16)
         | (ihex_int(in + 4) << 8)  |  ihex_int(in + 6);
    
    in += 8;
    
    // Verify checksum
    cksum = 4 + rtype
          + (uint8_t)((offset & 0xFF00) >> 8) + (uint8_t)(offset & 0x00FF)
          + (uint8_t)((addr & 0xFF000000) >> 24)
          + (uint8_t)((addr & 0x00FF0000) >> 16)
          + (uint8_t)((addr & 0x0000FF00) >> 8)
          + (uint8_t)(addr & 0x000000FF);
    
    cksum = (~cksum) + 1;
    
    if (cksum != ihex_int(in)) return 0xFFFFFFFF;
    
    return addr;
}


// Returns an encoded data record line from up to 255 bytes of input data
void ihex_encode_data (const char *in, char *out, uint8_t len, uint16_t offset) {
    
    // Let the raw line encoder do the heavy lifting
    ihex_encode_raw(in, out, len, offset, IHEX_RTYPE_DATA);
}


// Returns decoded data from an encoded data record line
uint8_t ihex_decode_data (const char *in, char *out, uint16_t *offset) {
    
    uint8_t len;
    uint8_t rtype;
    uint8_t cksum;
    uint8_t x;
    
    // Valid line?
    if (*(in++) != ':') return 0;
    
    // Get line length
    len = ihex_int(in);
    in += 2;
    
    // Get address offset
    *offset = (ihex_int(in) << 8) | (ihex_int(in + 2));
    in += 4;
    
    // Verify record type
    rtype = ihex_int(in);
    in += 2;
    
    if (rtype != IHEX_RTYPE_DATA) return 0;
    
    
    // Seed checksum with record preamble
    cksum = len + rtype
          + (uint8_t)((*offset & 0xFF00) >> 8) + (uint8_t)(*offset & 0x00FF);
    
    // Decode data bytes
    x = len;
    
    while (x) {
        
        // End loop on non-hex chars (inc. NULL, CR/LF, etc.)
        if ( (*in < '0') || (*in > 'f')  ||
            ((*in > '9') && (*in < 'A')) ||
            ((*in > 'F') && (*in < 'a')) ) return 0;
        
        // Convert hex digits to binary value
        *out = (char)ihex_int(in);
        in += 2;
        
        // Update checksum
        cksum += *((uint8_t*)out);
        
        out++;
        x--;
    }
    
    // Finally, verify the checksum
    cksum = (~cksum) + 1;
    
    if (cksum != ihex_int(in)) return 0;
    
    // Correctly decoded -- return data length
    return len;
}


// Generate checksum from (binary) input data
// NOTE: haven't found a use for this yet -- it might not stick around
uint8_t ihex_checksum (const char *in, uint8_t len) {
    
    uint8_t sum = 0;
    
    // Loop through input bytes, summing their values
    while (len) {
        sum += (uint8_t)*in;
        len--;
        in++;
    }
    
    return ((~sum) + 1);
}


// Write an 8-bit value to a char buffer as two hex digits
void ihex_hex (char *buf, uint8_t val) {
    
    uint8_t tmp;
    
    // High nibble
    tmp    = ((val & 0xF0) >> 4);
    buf[0] = (tmp > 9) ? tmp + 'A' - 10 : tmp + '0';
    
    // And low nibble
    tmp    = (val & 0x0F);
    buf[1] = (tmp > 9) ? tmp + 'A' - 10 : tmp + '0';
}


// Return 8-bit integer value from two hex chars
uint8_t ihex_int (const char *buf) {
    
    uint8_t tmp;
    
    // Check first digit (high nibble)
    if ( ((*buf) >= '0') && ((*buf) <= '9')) {
        tmp = ((*buf) - '0') << 4;
    } else if ( ((*buf) >= 'A') && ((*buf) <= 'F') ) {
        tmp = ((*buf) + 10 - 'A') << 4;
    } else if ( ((*buf) >= 'a') && ((*buf) <= 'f') ) {
        tmp = ((*buf) + 10 - 'a') << 4;
    } else {
        return 0;       // Error
    }
    
    buf++;
    
    // Check second digit (low nibble)
    if ( ((*buf) >= '0') && ((*buf) <= '9')) {
        tmp |= (*buf) - '0';
    } else if ( ((*buf) >= 'A') && ((*buf) <= 'F') ) {
        tmp |= (*buf) + 10 - 'A';
    } else if ( ((*buf) >= 'a') && ((*buf) <= 'f') ) {
        tmp |= (*buf) + 10 - 'a';
    } else {
        return 0;       // Error
    }
    
    return tmp;
}