#include "ihex-file.h"

int main (int argc, char *argv[]) {
    
    int      retval;            // Function return value
    
    char    *infile = NULL;     // Filename for input file
    char    *hexfile = NULL;    // Filename for hex output file
    char    *outfile = NULL;    // Filename for binary output file
    FILE    *fp_in  = NULL;     // File pointer for input file
    FILE    *fp_hex = NULL;     // File pointer for output hex file
    FILE    *fp_out = NULL;     // File pointer for output bin file
    
    char    *buf1 = NULL;       // Buffer for file I/O (input)
    char    *buf2 = NULL;       // Buffer for file I/O (output, or 2nd input)
    size_t   flen1;             // Return value - bytes read/written
    size_t   flen2;             // Return value - bytes read/written
    uint32_t offset;            // Current memory offset, or file pos
    uint16_t addr;              // Segment, address part, etc..
    uint32_t data;              // Spare data
    
    
    // Check CLI
    if (argc < 1) {
        fprintf(stderr, "%s: No input file specified\n", argv[0]);
        goto Error;
    }
    
    
    // Allocate memory for filenames and buffers
    hexfile = malloc(strlen(TMPLATE) + 1);
    
    if (hexfile == NULL) {
        fprintf(stderr, "Failed to allocate memory for filename template\n");
        goto Error;
    }
    
    outfile = malloc(strlen(TMPLATE) + 1);
    
    if (outfile == NULL) {
        fprintf(stderr, "Failed to allocate memory for filename template\n");
        goto Error;
    }
    
    buf1 = malloc(BUFLEN);
    
    if (buf1 == NULL) {
        fprintf(stderr, "Failed to allocate buffer memory\n");
        goto Error;
    }
    
    buf2 = malloc(LINELEN);
    
    if (buf2 == NULL) {
        fprintf(stderr, "Failed to allocate buffer memory\n");
        goto Error;
    }
    
    
    // Open input file
    infile = argv[1];
    
    printf("Opening %s...\n", infile);
    
    fp_in = fopen(infile, "rb");
    
    if (fp_in == NULL) {
        fprintf( stderr, "Failed to open file: (%i) %s\n",
                 errno, strerror(errno) );
        goto Error;
    }
    
    
    // Create a temp file for hex output
    printf("Creating hex output file...");
    fflush(stdout);
    
    strcpy(hexfile, TMPLATE);
    
    retval = mkstemp(hexfile);
    
    if (retval < 0) {
        fprintf( stderr, "Failed to create temp file '%s': (%i) %s\n",
                 hexfile, errno, strerror(errno) );
        goto Error;
    }
    
    fp_hex = fdopen(retval, "w+b");
    
    if (fp_hex == NULL) {
        
        fprintf( stderr, "Failed to open temp file '%s': (%i) %s\n",
                 hexfile, errno, strerror(errno) );
        goto Error;
    }
    
    printf("'%s' created\n", hexfile);
    
    
    // Create another temp file for binary output
    printf("Creating binary output file... ");
    fflush(stdout);
    
    strcpy(outfile, TMPLATE);
    
    retval = mkstemp(outfile);
    
    if (retval < 0) {
        fprintf( stderr, "Failed to create temp file '%s': (%i) %s\n",
                 outfile, errno, strerror(errno) );
        goto Error;
    }
    
    fp_out = fdopen(retval, "w+b");
    
    if (fp_out == NULL) {
        fprintf( stderr, "Failed to open temp file '%s': (%i) %s\n",
                 outfile, errno, strerror(errno) );
        goto Error;
    }
    
    printf("'%s' created\n", outfile);
    
    
    // Encode binary input file to Intel Hex format
    errno  = 0;
    addr   = 0 - 1;
    offset = 0;
    
    printf("Encoding Intel Hex file...\n");
    
    while (errno == 0) {
        
        // Don't cross segment (high-word) boundaries
        flen2 = (0x10000 - (offset & 0x0000FFFF));
        
        if (flen2 > BUFLEN) flen2 = BUFLEN;
        
        flen1 = fread(buf1, 1, flen2, fp_in);
        
        if (flen1 < 1) break;
        
        // Write new 32-bit address record when 16-bit offset has overflowed
        if (addr != ((offset & 0xFFFF0000) >> 16)) {
            
            addr = ((offset & 0xFFFF0000) >> 16);
            
            ihex_encode_ext_addr(buf2, IHEX_RTYPE_EXTADDR32, addr);
            
            fprintf(fp_hex, "%s\n", buf2);
        }
        
        ihex_encode_data(buf1, buf2, flen1, (offset & 0x0000FFFF));
        
        fprintf(fp_hex, "%s\n", buf2);
        
        offset += flen1;
    }
    
    if (errno != 0) {
        fprintf(stderr, "I/O error: (%i) %s\n", errno, strerror(errno));
        goto Error;
    }
    
    // Write EOF record
    printf("Writing EOF...\n");
    
    ihex_encode_eof(buf2);
    
    fprintf(fp_hex, "%s\n", buf2);
    
    
    // Decode Intel Hex file back to binary
    errno  = 0;
    addr   = 0;
    offset = 0;
    
    printf("Decoding Intel Hex file...\n");
    
    fseek(fp_hex, 0, SEEK_SET);
    
    while (errno == 0) {
        
        flen1 = 0;
        
        while (flen1 <= IHEX_LINEMAX) {
            
            retval = getc(fp_hex);
            
            // Error, or end-of-line
            if ( (retval < 0) || (retval == 0x0d) || (retval == 0x0a) )
                break;
            
            // Got a character
            *(buf2 + flen1) = (char)retval;
            
            flen1++;
        }
        
        // Empty line, keep going
        if (flen1 == 0) continue;
        
        // Incomplete line, bail out
        if (flen1 < (IHEX_LINEMIN - 1)) {
            fprintf( stderr, "Line length below minimum valid: %zu of %u\n",
                     flen1, (IHEX_LINEMIN - 1) );
            break;
        }
        
        // Exceeded buffer length
        if ((flen1 == IHEX_LINEMAX) && (*(buf2 + flen1) != 0x00)) {
            fprintf(stderr, "Line exceeded buffer length (%u): '", IHEX_LINEMAX);
            fwrite(buf2, 1, IHEX_LINEMAX, stderr);
            fprintf(stderr, "'\n");
            break;
        }
        
        // See what kind of record was found
        retval = ihex_get_record_type(buf2);
        if (retval == IHEX_RTYPE_INVALID) {
            fprintf(stderr, "Invalid line: '%s'\n", buf2);
            break;
        }
        
        // End of file
        if (retval == IHEX_RTYPE_EOF) break;
        
        // 16-bit segment address
        if (retval == IHEX_RTYPE_EXTADDR16) {
            offset = ihex_decode_ext_addr(buf2) * 16;
            printf("Got segment address: 0x%08x\n", offset);
            continue;
        }
        
        // 32-bit linear address
        if (retval == IHEX_RTYPE_EXTADDR32) {
            offset = ihex_decode_ext_addr(buf2) << 16;
            printf("Got linear address: 0x%08x\n", offset);
            continue;
        }
        
        // CS:IP record
        if (retval == IHEX_RTYPE_STARTADDR16) {
            printf( "Got Start Address (CS:IP): 0x%08x",
                    ihex_decode_start_addr(buf2) );
            continue;
        }
        
        // EIP record
        if (retval == IHEX_RTYPE_STARTADDR32) {
            printf( "Got Start Address (EIP): 0x%08x",
                    ihex_decode_start_addr(buf2) );
            continue;
        }
        
        // Unsupported record type
        if (retval != IHEX_RTYPE_DATA) {
            fprintf(stderr, "Unknown record type: %i\n", retval);
            break;
        }
        
        // Data record
        retval = ihex_decode_data(buf2, buf1, &addr);
        
        if (!retval) {
            fprintf(stderr, "Failed to decode line, or no data: '%s'\n", buf2);
            break;
        }
        
        flen1 = retval;
        
        // Check address
        if (addr != (offset & 0x0000FFFF)) {
            fprintf( stderr,
                     "Address does not match: expected 0x%04x, got 0x%04x\n",
                     (offset & 0x0000FFFF), addr );
            break;
        }
        
        // Update address
        offset += flen1;
        
        
        // Write decoded data to binary output file
        flen2 = fwrite(buf1, 1, flen1, fp_out);
        
        if (flen2 < flen1) {
            fprintf( stderr, "Short write: expected %zu, got %zu\n",
                     flen1, flen2 );
            break;
        }
    }
    
    if (errno != 0) {
        fprintf(stderr, "I/O error: (%i) %s\n", errno, strerror(errno));
        goto Error;
    }
    
    
    // Compare resulting binary with original file
    errno  = 0;
    offset = 0;
    
    printf("Comparing decoded file to original...\n");
    
    fseek(fp_in, 0, SEEK_SET);
    fseek(fp_out, 0, SEEK_SET);
    
    while (errno == 0) {
        
        flen1 = fread(buf1, 1, BUFLEN, fp_in);
        
        if (flen1 < 1) break;
        
        flen2 = fread(buf2, 1, BUFLEN, fp_out);
        
        if (flen2 < 1) break;
        
        
        // Compare read lengths
        if (flen1 != flen2) {
            fprintf( stderr, "Read mismatch: read %zu bytes from input file, "
                     "%zu bytes from output file\n", flen1, flen2);
            goto Error;
        }
        
        // Compare buffers
        if (memcmp(buf1, buf2, flen1) != 0) {
            fprintf( stderr, "Read mismatch: buffer contents differ reading "
                     "%zu bytes at offset 0x%08x\n", flen1, offset);
            goto Error;
        }
        
        // Update offset counter
        offset += flen1;
    }
    
    if (errno != 0) {
        fprintf(stderr, "I/O error: (%i) %s\n", errno, strerror(errno));
        goto Error;
    }
    
    printf("Test passed!  Output file matches original input.\n");
    
    
    // Close files
    printf("Closing files...\n");
    
    fclose(fp_in);
    fclose(fp_hex);
    fclose(fp_out);
    
    
    // Remove temp files
    printf("Removing temp files...\n");
    
    unlink(hexfile);
    unlink(outfile);
    
    
    // Free memory
    printf("Freeing memory...\n");
    
    free(hexfile);
    free(outfile);
    
    free(buf1);
    free(buf2);
    
    
    return 0;
    
    
    // Error handler
    Error:
    
    if (fp_in != NULL) {
        fclose(fp_in);
    }
    
    if (fp_hex != NULL) {
        fclose(fp_hex);
        //unlink(hexfile);      // Keep this for analysis
    }
    
    if (fp_out != NULL) {
        fclose(fp_out);
        //unlink(outfile);      // Keep this for analysis
    }
    
    if (hexfile != NULL) free(hexfile);
    if (outfile != NULL) free(outfile);
    
    if (buf1 != NULL) free(buf1);
    if (buf2 != NULL) free(buf2);
    
    return 1;
}