#include <inttypes.h>
#include <malloc.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>

#include "intelhex.h"

struct rtype_t {
    uint8_t  rtype;
    uint8_t  checksum;
    char     sample[64];
};

int main (int argc, char *argv[]) {
    
    uint8_t  i = 0;
    uint8_t  rtype;
    uint16_t addr16;
    uint32_t addr32;
    
    char     buf[256];
    
    
    // Known-good sample records
    struct rtype_t  rtypes[] = {
        { .rtype    = IHEX_RTYPE_DATA,
          .checksum = 0x86,
          .sample   = ":0c01230048656C6C6f20576f726c642193" },
        { .rtype    = IHEX_RTYPE_EOF,
          .checksum = 0xFF,
          .sample   = ":00000001FF" },
        { .rtype    = IHEX_RTYPE_EXTADDR16,
          .checksum = 0xE7,
          .sample   = ":020000020a0bE7" },
        { .rtype    = IHEX_RTYPE_STARTADDR16,
          .checksum = 0xCB,
          .sample   = ":040000030a0b0c0dCB" },
        { .rtype    = IHEX_RTYPE_EXTADDR32,
          .checksum = 0xAA,
          .sample   = ":02000004a0b0AA" },
        { .rtype    = IHEX_RTYPE_STARTADDR32,
          .checksum = 0x17,
          .sample   = ":04000005a0b0c0d017" },
        { .rtype    = IHEX_RTYPE_INVALID,
          .checksum = 0x00,
          .sample   = ":0000001000" },
        { .rtype    = 0x00,
          .checksum = 0x00,
          .sample   = "" }
    };
    
    
    printf("Testing record type parser...\n");
    
    while (rtypes[i].sample[0] != 0x00) {
        
        rtype = ihex_get_record_type(rtypes[i].sample);
        
        if (rtype != rtypes[i].rtype) {
            fprintf( stderr, "Record %u: Expected rtype 0x%02x, got 0x%02x\n",
                     i, rtypes[i].rtype, rtype );
            return 1;
        }
        
        i++;
    };
    
    
    printf("Testing encoders...\n");
    
    // EOF
    ihex_encode_eof(buf);
    
    if (strcasecmp(buf, rtypes[IHEX_RTYPE_EOF].sample) != 0) {
        fprintf( stderr, "EOF record encoder failed; expected '%s', got '%s'\n",
                 rtypes[IHEX_RTYPE_EOF].sample, buf );
        return 1;
    }
    
    // Data
    ihex_encode_data("Hello World!", buf, 12, 0x0123);
    
    if (strcasecmp(buf, rtypes[IHEX_RTYPE_DATA].sample) != 0) {
        fprintf( stderr, "Data record encoder failed; "
                 "expected '%s', got '%s'\n",
                 rtypes[IHEX_RTYPE_DATA].sample, buf );
        return 1;
    }
    
    // Extended Segment Address
    ihex_encode_ext_addr(buf, IHEX_RTYPE_EXTADDR16, 0x0a0b);
    
    if (strcasecmp(buf, rtypes[IHEX_RTYPE_EXTADDR16].sample) != 0) {
        fprintf( stderr,
                 "Extended Segment Address encoder failed; "
                 "expected '%s', got '%s'\n",
                 rtypes[IHEX_RTYPE_EXTADDR16].sample, buf );
        return 1;
    }
    
    // Extended Linear Address
    ihex_encode_ext_addr(buf, IHEX_RTYPE_EXTADDR32, 0xa0b0);
    
    if (strcasecmp(buf, rtypes[IHEX_RTYPE_EXTADDR32].sample) != 0) {
        fprintf( stderr,
                 "Extended Linear Address encoder failed; "
                 "expected '%s', got '%s'\n",
                 rtypes[IHEX_RTYPE_EXTADDR32].sample, buf );
        return 1;
    }
    
    // Start Segment Address
    ihex_encode_start_addr(buf, IHEX_RTYPE_STARTADDR16, 0x0a0b0c0d);
    
    if (strcasecmp(buf, rtypes[IHEX_RTYPE_STARTADDR16].sample) != 0) {
        fprintf( stderr,
                 "Start Segment Address encoder failed; "
                 "expected '%s', got '%s'\n",
                 rtypes[IHEX_RTYPE_STARTADDR16].sample, buf );
        return 1;
    }
    
    // Start Linear Address
    ihex_encode_start_addr(buf, IHEX_RTYPE_STARTADDR32, 0xa0b0c0d0);
    
    if (strcasecmp(buf, rtypes[IHEX_RTYPE_STARTADDR32].sample) != 0) {
        fprintf( stderr,
                 "Start Linear Address encoder failed; "
                 "expected '%s', got '%s'\n",
                 rtypes[IHEX_RTYPE_STARTADDR32].sample, buf );
        return 1;
    }
    
    
    printf("Testing decoders...\n");
    
    // Data
    i = ihex_decode_data(rtypes[IHEX_RTYPE_DATA].sample, buf, &addr16);
    
    buf[i] = 0x00;  // Add string terminator at end-of-data
    
    if ( (strcmp(buf, "Hello World!") != 0) || (addr16 != 0x0123) ) {
        fprintf( stderr,
                 "Data record decoder failed; "
                 "expected data: '%s', address: 0x%04x; got '%s', 0x%04x\n",
                 "Hello World!", 0x0123, buf, addr16 );
        return 1;
    }
    
    // Extended Segment Address
    addr16 = ihex_decode_ext_addr(rtypes[IHEX_RTYPE_EXTADDR16].sample);
    
    if (addr16 != 0x0a0b) {
        fprintf( stderr,
                 "Extended Segment Address decoder failed; "
                 "expected '0x%04x', got '0x%04x'\n",
                 0x0a0b, addr16 );
        return 1;
    }
    
    // Extended Linear Address
    addr16 = ihex_decode_ext_addr(rtypes[IHEX_RTYPE_EXTADDR32].sample);
    
    if (addr16 != 0xa0b0) {
        fprintf( stderr,
                 "Extended Linear Address decoder failed; "
                 "expected '0x%04x', got '0x%04x'\n",
                 0xa0b0, addr16 );
        return 1;
    }
    
    // Start Segment Address
    addr32 = ihex_decode_start_addr(rtypes[IHEX_RTYPE_STARTADDR16].sample);
    
    if (addr32 != 0x0a0b0c0d) {
        fprintf( stderr,
                 "Start Segment Address decoder failed; "
                 "expected '0x%08x', got '0x%08x'\n",
                 0x0a0b0c0d, addr32 );
        return 1;
    }
    
    // Start Linear Address
    addr32 = ihex_decode_start_addr(rtypes[IHEX_RTYPE_STARTADDR32].sample);
    
    if (addr32 != 0xa0b0c0d0) {
        fprintf( stderr,
                 "Start Linear Address decoder failed; "
                 "expected '0x%08x', got '0x%08x'\n",
                 0xa0b0c0d0, addr32 );
        return 1;
    }
    
    
    printf("All tests passed!\n");
    
    return 0;
}