#include "uu-file.h"

int main (int argc, char *argv[]) {
    
    int      retval;            // Function return value
    
    char    *infile = NULL;     // Filename for input file
    char    *uufile = NULL;     // Filename for UUencoded output file
    char    *outfile = NULL;    // Filename for binary output file
    FILE    *fp_in  = NULL;     // File pointer for input file
    FILE    *fp_uu = NULL;      // File pointer for UUencoded output file
    FILE    *fp_out = NULL;     // File pointer for binary output file
    
    char    *buf1 = NULL;       // Buffer for file I/O (input)
    char    *buf2 = NULL;       // Buffer for file I/O (output, or 2nd input)
    size_t   flen1;             // Return value - bytes read/written
    size_t   flen2;             // Return value - bytes read/written
    
    uint32_t offset;            // Current file pos
    uint32_t cksum;             // Checksum
    
    
    // Check CLI
    if (argc < 1) {
        fprintf(stderr, "%s: No input file specified\n", argv[0]);
        goto Error;
    }
    
    
    // Allocate memory for filenames and buffers
    uufile = malloc(strlen(TMPLATE) + 1);
    
    if (uufile == NULL) {
        fprintf(stderr, "Failed to allocate memory for filename template\n");
        goto Error;
    }
    
    outfile = malloc(strlen(TMPLATE) + 1);
    
    if (outfile == NULL) {
        fprintf(stderr, "Failed to allocate memory for filename template\n");
        goto Error;
    }
    
    buf1 = malloc(UU_DATAMAX);
    
    if (buf1 == NULL) {
        fprintf(stderr, "Failed to allocate buffer memory\n");
        goto Error;
    }
    
    buf2 = malloc(UU_LINEMAX);
    
    if (buf2 == NULL) {
        fprintf(stderr, "Failed to allocate buffer memory\n");
        goto Error;
    }
    
    
    // Open input file
    infile = argv[1];
    
    printf("Opening %s...\n", infile);
    
    fp_in = fopen(infile, "rb");
    
    if (fp_in == NULL) {
        fprintf( stderr, "Failed to open file: (%i) %s\n",
                 errno, strerror(errno) );
        goto Error;
    }
    
    
    // Create a temp file for UUencoded output
    printf("Creating UUencoded output file...");
    fflush(stdout);
    
    strcpy(uufile, TMPLATE);
    
    retval = mkstemp(uufile);
    
    if (retval < 0) {
        fprintf( stderr, "Failed to create temp file '%s': (%i) %s\n",
                 uufile, errno, strerror(errno) );
        goto Error;
    }
    
    fp_uu = fdopen(retval, "w+b");
    
    if (fp_uu == NULL) {
        
        fprintf( stderr, "Failed to open temp file '%s': (%i) %s\n",
                 uufile, errno, strerror(errno) );
        goto Error;
    }
    
    printf("'%s' created\n", uufile);
    
    
    // Create another temp file for binary output
    printf("Creating binary output file... ");
    fflush(stdout);
    
    strcpy(outfile, TMPLATE);
    
    retval = mkstemp(outfile);
    
    if (retval < 0) {
        fprintf( stderr, "Failed to create temp file '%s': (%i) %s\n",
                 outfile, errno, strerror(errno) );
        goto Error;
    }
    
    fp_out = fdopen(retval, "w+b");
    
    if (fp_out == NULL) {
        fprintf( stderr, "Failed to open temp file '%s': (%i) %s\n",
                 outfile, errno, strerror(errno) );
        goto Error;
    }
    
    printf("'%s' created\n", outfile);
    
    
    // UUencode binary input file
    errno = 0;
    cksum = 0;
    
    printf("UUencoding input file...\n");
    
    while (errno == 0) {
        
        flen1 = fread(buf1, 1, UU_DATAMAX, fp_in);
        
        if (flen1 < 1) break;
        
        uu_encode_line(buf1, buf2, flen1, &cksum);
        
        fprintf(fp_uu, "%s\n", buf2);
    }
    
    if (errno != 0) {
        fprintf(stderr, "I/O error: (%i) %s\n", errno, strerror(errno));
        goto Error;
    }
    
    // Write checksum
    printf("Writing checksum...\n");
    
    fprintf(fp_uu, "%u\n", cksum);
    
    
    // Decode back to binary
    errno  = 0;
    cksum  = 0;
    offset = 0;
    
    printf("Decoding UUencoded file...\n");
    
    fseek(fp_uu, 0, SEEK_SET);
    
    while (errno == 0) {
        
        flen1 = 0;
        
        while (flen1 <= UU_LINEMAX) {
            
            retval = getc(fp_uu);
            
            // Error, or end-of-line
            if ( (retval < 0) || (retval == 0x0d) || (retval == 0x0a) )
                break;
            
            // Got a character
            *(buf2 + flen1) = (char)retval;
            
            flen1++;
        }
        
        // EOF, get outta here
        if (feof(fp_uu)) break;
        
        // Empty line, keep going
        if (flen1 == 0) continue;
        
        // Exceeded buffer length
        if ((flen1 == UU_LINEMAX) && (*(buf2 + flen1) != 0x00)) {
            
            fprintf(stderr, "Line exceeded buffer length (%u): '", UU_LINEMAX);
            fwrite(buf2, 1, UU_LINEMAX, stderr);
            fprintf(stderr, "'\n");
            
            goto Error;
        }
        
        
        // Look for solitary checksum line using heuristics,
        // since we don't know the total payload size
        // (Forgive the sloppy re-use of exisitng variables)
        
        retval = 1;     // Flag: 0=Just data; 1=Could be a checksum
        
        for (flen2 = 0; flen2 < flen1; flen2++) {
            
            // Non-numerical digit, so this is a normal data line
            if ( (buf2[flen2] < '0') || (buf2[flen2] > '9') )
                retval = 0;
        }
        
        
        // Compare checksums
        if (retval) {
            
            // Convert derived checksum to a string
            sprintf(buf1, "%"PRIu32, cksum);
            
            // Null-terminate checksum read from file
            buf2[flen1] = 0x00;
            
            // Compare as strings
            if (strcmp(buf1, buf2) != 0) {
                
                fprintf( stderr, "Checksums differ: Got '%s', expected '%s'\n",
                         buf2, buf1 );
                
                goto Error;
            }
            
            // Matched
            printf("Got checksum: 0x%08x (%s)\n", cksum, buf2);
            break;
        }
        
        
        // Data line
        retval = uu_decode_line(buf2, buf1, &cksum);
        
        if (!retval) {
            fprintf(stderr, "Failed to decode line, or no data: '%s'\n", buf2);
            break;
        }
        
        flen1 = retval;
        
        
        // Write decoded data to binary output file
        flen2 = fwrite(buf1, 1, flen1, fp_out);
        
        if (flen2 < flen1) {
            fprintf( stderr, "Short write: expected %zu, got %zu\n",
                     flen1, flen2 );
            break;
        }
    }
    
    if (errno != 0) {
        fprintf(stderr, "I/O error: (%i) %s\n", errno, strerror(errno));
        goto Error;
    }
    
    
    // Compare resulting binary with original file
    errno  = 0;
    offset = 0;
    
    printf("Comparing decoded file to original...\n");
    
    fseek(fp_in, 0, SEEK_SET);
    fseek(fp_out, 0, SEEK_SET);
    
    while (errno == 0) {
        
        flen1 = fread(buf1, 1, UU_DATAMAX, fp_in);
        
        if (flen1 < 1) break;
        
        flen2 = fread(buf2, 1, UU_DATAMAX, fp_out);
        
        if (flen2 < 1) break;
        
        
        // Compare read lengths
        if (flen1 != flen2) {
            fprintf( stderr, "Read mismatch: read %zu bytes from input file, "
                     "%zu bytes from output file\n", flen1, flen2);
            goto Error;
        }
        
        // Compare buffers
        if (memcmp(buf1, buf2, flen1) != 0) {
            fprintf( stderr, "Read mismatch: buffer contents differ reading "
                     "%zu bytes at offset 0x%08x\n", flen1, offset);
            goto Error;
        }
        
        // Update offset counter
        offset += flen1;
    }
    
    if (errno != 0) {
        fprintf(stderr, "I/O error: (%i) %s\n", errno, strerror(errno));
        goto Error;
    }
    
    printf("Test passed!  Output file matches original input.\n");
    
    
    // Close files
    printf("Closing files...\n");
    
    fclose(fp_in);
    fclose(fp_uu);
    fclose(fp_out);
    
    
    // Remove temp files
    printf("Removing temp files...\n");
    
    unlink(uufile);
    unlink(outfile);
    
    
    // Free memory
    printf("Freeing memory...\n");
    
    free(uufile);
    free(outfile);
    
    free(buf1);
    free(buf2);
    
    
    return 0;
    
    
    // Error handler
    Error:
    
    if (fp_in != NULL) {
        fclose(fp_in);
    }
    
    if (fp_uu != NULL) {
        fclose(fp_uu);
        //unlink(uufile);       // Keep this for analysis
    }
    
    if (fp_out != NULL) {
        fclose(fp_out);
        //unlink(outfile);      // Keep this for analysis
    }
    
    if (uufile  != NULL) free(uufile);
    if (outfile != NULL) free(outfile);
    
    if (buf1 != NULL) free(buf1);
    if (buf2 != NULL) free(buf2);
    
    return 1;
}