/*
 * Serial Test -- POSIX implementation of PC-to-IC ISP flash
 *
 * unix-cmds.c
 *
 * This file contains the boot ROM command interface library.  All functions
 * that talk to the boot ROM are implemented here.
 */

#include "unix.h"


// Globals defined in unix.c:
extern int      verbose;
extern int      quiet;
extern uint16_t timeout;


// Consume input until a given string is found
int waitfor (int fd, char *data) {
    
    char     buf;           // Character buffer
    unsigned bpos = 0;      // Match data buffer position
    
    int      retval;        // Function return value
    uint16_t cnt = timeout; // Timeout counter
    
    
    // Read in non-blocking mode
    fcntl(fd, F_SETFL, FNDELAY);
    
    // Loop until end of data (NULL) or timeout
    while (*(data + bpos) != 0x00) {
        
        // Check for incoming data
        retval = read(fd, &buf, 1);
        
        // Got End-of-File
        if (retval == 0) {
            errno = ENODATA;
            return -1;
        }
        
        // Got a byte -- compare to input
        if (retval == 1) {
            
            // Character matches -- move on to next byte
            if ((char)*(data + bpos) == buf) {
                bpos++;
                continue;
            }
            
            // Character did not match, reset match buffer position
            bpos = 0;
            continue;
        }
        
        // Got an error
        if (retval < 0) {
            
            // No data pending
            if ( (errno == EAGAIN) || (errno == EWOULDBLOCK) ) {
                
                // Timed out
                if (cnt == 0) {
                    errno = ETIMEDOUT;
                    return 0;
                }
                
                // Sleep for a ms and try again
                usleep(1000);
                cnt--;
                continue;
            }
            
            // Something more serious happened
            return -1;
        }
    } // while
    
    
    // Data matches!
    return 1;
}


// Read a line terminated by CRLF
int readline (int fd, char *buf, unsigned buflen) {
    
    int      retval;        // Function return value
    uint16_t cnt = timeout; // Timeout counter
    
    unsigned bpos = 0;      // Buffer position
    char     val;           // Incoming character
    
    buf[buflen-1] = 0x00;
    // Read in non-blocking mode
    fcntl(fd, F_SETFL, FNDELAY);
    
    while (bpos < buflen) {
        
        // Check for incoming data
        retval = read(fd, &val, 1);
        
        // Got End-of-File
        if (retval == 0) {
            errno = ENODATA;
            return -1;
        }
        
        // Got an error
        if (retval < 0) {
            
            // No data available yet
            if ( (errno == EAGAIN) || (errno == EWOULDBLOCK) ) {
                
                // Timed out
                if (cnt == 0) {
                    errno = ETIMEDOUT;
                    return 0;
                }
                
                // Sleep for 1ms and try again
                usleep(1000);
                cnt--;
                continue;
            }
            
            // Something more serious happened
            return -1;
        }
        
        // Terminate line on CR, and terminate string with a NULL
        if (val == 0x0d) {
            buf[bpos] = 0x00;
            return 1;
        }
        
        // Ignore LF (might be left over from previous line)
        if (val == 0x0a) continue;
        
        // Not end-of-line yet, copy character to buffer and carry on
        buf[bpos] = val;
        bpos++;
    }
    
    errno = EOVERFLOW;
    return 0;
}


// Probe device and attempt to establish synchronization
int synchronize (int fd) {
    
    int retval;             // Function return value
    
    // Send question marks until the device responds
    while (1) {
        
        // Block on writes
        fcntl(fd, F_SETFL, 0);
        
        retval = write(fd, "?", 1);
        
        if (retval < 1) return -1;
        
        retval = waitfor(fd, "Synchronized");
        
        if (retval < 0) return 0;
        
        // Found a match!
        if (retval) break;
        
        // Timed out
        errno = ETIMEDOUT;
        return 0;
    }
    
    
    // Respond back to device
    fcntl(fd, F_SETFL, 0);
    
    retval = write(fd, "Synchronized\r\n", 14);
    
    if (retval < 1) return 0;
    
    retval = waitfor(fd, "OK\r\n");
    
    // Found a match!
    if (retval) return 1;
    
    // Error or timeout
    return 0;
}


// Inform device of its actual clock frequency
int set_clock (int fd, uint32_t clock) {
    
    int  retval;            // Function return value
    char buf[BUFLEN];       // Clock speed as string
    
    
    // Convert clock speed to string
    retval = snprintf(buf, BUFLEN, "%"PRIu32, clock);
    
    // Conversion failed
    if (retval < 0) return 0;
    
    // Value (possibly) too large
    if (retval >= (BUFLEN - 1)) {
        errno = EOVERFLOW;
        return 0;
    }
    
    // Add \r\n and a new null
    strcpy(buf + strlen(buf), "\r\n");
    
    // Set blocking mode again
    fcntl(fd, F_SETFL, 0);
    
    retval = write(fd, buf, strlen(buf));
    
    if (retval < 1) return 0;
    
    // Consume one line (the echo from our write)
    retval = readline(fd, buf, BUFLEN);
    
    // Time out, no data, buffer overrun, or other error
    if (retval < 1) return 0;
    
    // Now check for the MCU's actual response
    retval = readline(fd, buf, BUFLEN);
    
    // Time out, no data, buffer overrun, or other error
    if (retval < 1) return 0;
        
    // Clock speed accepted by the MCU?
    if (strcmp(buf, "OK") == 0) return 1;
    
    // Got an unexpected result
    if (verbose) fprintf(stderr, "MCU returned: '%s'\n", buf);
    
    errno = EINVAL;
    
    return 0;
}


// Turn off serial echo
int echo_off (int fd) {
    
    int  retval;            // Function return value
    char buf[BUFLEN];       // Response buffer
    
    // Build command string
    strcpy(buf, "A 0\r\n");
    
    // Set blocking mode
    fcntl(fd, F_SETFL, 0);
    
    retval = write(fd, buf, strlen(buf));
    
    if (retval < 1) return 0;
    
    // Depending on whether echo was actually enabled, we could get:
    // an echo response, "0" (CMD_SUCCESS), or something else
    while (1) {
        
        // Fetch a line
        retval = readline(fd, buf, BUFLEN);
        
        if (retval < 1) return 0;
        
        // If we got an echo, consume it
        if (strcmp(buf, "A 0") == 0) continue;
        
        // If we got 0, we're done
        if (strcmp(buf, "0") == 0) return 1;
        
        // Something else?  Uh oh...
        break;
    }
    
    if (verbose) fprintf(stderr, "MCU returned: '%s'\n", buf);
    
    errno = EINVAL;
    
    return 0;
}


// Ask device to reconfigure its UART for a new speed
int set_new_speed (int fd, uint32_t baud, uint8_t stop) {
    
    int  retval;            // Function return value
    char buf[BUFLEN];       // Response buffer
    
    // Build command string
    retval = snprintf(buf, BUFLEN, "B %"PRIu32" %u\r\n", baud, stop);
    
    // Set blocking mode
    fcntl(fd, F_SETFL, 0);
    
    retval = write(fd, buf, strlen(buf));
    
    if (retval < 1) return 0;
    
    retval = readline(fd, buf, BUFLEN);
    
    if (retval < 1) return 0;
    
    // Check for affirmative response
    if (strcmp(buf, "0") == 0) return 1;
    
    // Error
    if (verbose) fprintf(stderr, "MCU returned: '%s'\n", buf);
    
    errno = EINVAL;
    
    return 0;
}


// Query device part ID
int get_part_id (int fd, uint32_t *id) {
    
    int  retval;            // Function return value
    char buf[BUFLEN];       // Response buffer
    
    // Build command string
    strcpy(buf, "J\r\n");
    
    // Set blocking mode
    fcntl(fd, F_SETFL, 0);
    
    retval = write(fd, buf, strlen(buf));
    
    if (retval < 1) return 0;
    
    retval = readline(fd, buf, BUFLEN);
    
    if (retval < 1) return 0;
    
    // Check for response
    if (strcmp(buf, "0") != 0) {
        if (verbose) fprintf(stderr, "MCU returned: '%s'\n", buf);
        errno = EINVAL;
        return 0;
    }
    
    // Get part ID
    retval = readline(fd, buf, BUFLEN);
    
    if (retval < 1) return 0;
    
    sscanf(buf, "%"SCNu32, id);
    
    return 1;
}


// Query device unique ID
int get_unique_id (int fd, uint32_t *uid) {
    
    int     retval;         // Function return value
    char    buf[BUFLEN];    // Response buffer
    uint8_t i;              // Loop counter
    
    // Build command string
    strcpy(buf, "N\r\n");
    
    // Set blocking mode
    fcntl(fd, F_SETFL, 0);
    
    retval = write(fd, buf, strlen(buf));
    
    if (retval < 1) return 0;
    
    retval = readline(fd, buf, BUFLEN);
    
    if (retval < 1) return 0;
    
    // Check for response
    if (strcmp(buf, "0") != 0) {
        if (verbose) fprintf(stderr, "MCU returned: '%s'\n", buf);
        errno = EINVAL;
        return 0;
    }
    
    // Get unique ID
    for (i = 0; i < 4; i++) {
        
        retval = readline(fd, buf, BUFLEN);
        
        if (retval < 1) return 0;
        
        sscanf(buf, "%"SCNu32, &uid[i]);
    }
    
    return 1;
}


// Query device boot ROM version
int get_rom_version (int fd, uint8_t *ver) {
    
    int     retval;         // Function return value
    char    buf[BUFLEN];    // Response buffer
    uint8_t i;              // Loop counter
    
    // Build command string
    strcpy(buf, "K\r\n");
    
    // Set blocking mode
    fcntl(fd, F_SETFL, 0);
    
    retval = write(fd, buf, strlen(buf));
    
    if (retval < 1) return 0;
    
    retval = readline(fd, buf, BUFLEN);
    
    if (retval < 1) return 0;
    
    // Check for response
    if (strcmp(buf, "0") != 0) {
        if (verbose) fprintf(stderr, "MCU returned: '%s'\n", buf);
        errno = EINVAL;
        return 0;
    }
    
    // Get unique ID
    for (i = 0; i < 2; i++) {
        
        retval = readline(fd, buf, BUFLEN);
        
        if (retval < 1) return 0;
        
        sscanf(buf, "%"SCNu8, &ver[i]);
    }
    
    return 1;
}


// Unlock flash access
int unlock (int fd, uint32_t code) {
    
    int     retval;         // Function return value
    char    buf[BUFLEN];    // Response buffer
    
    // Build command string
    retval = snprintf(buf, BUFLEN, "U %"PRIu32"\r\n", code);
    
    // Set blocking mode
    fcntl(fd, F_SETFL, 0);
    
    retval = write(fd, buf, strlen(buf));
    
    if (retval < 1) return 0;
    
    retval = readline(fd, buf, BUFLEN);
    
    if (retval < 1) return 0;
    
    // Check for response
    if (strcmp(buf, "0") != 0) {
        if (verbose) fprintf(stderr, "MCU returned: '%s'\n", buf);
        errno = EINVAL;
        return 0;
    }
    
    return 1;
}


// Check flash sectors to see if they're all blank
int check_blank (int fd, uint8_t start, uint8_t len) {
    
}