/* 
 * Serial Test -- POSIX implementation of PC-to-IC ISP flash
 * 
 * unix.c
 * 
 * This file contains the entry point, command-line processing, serial setup,
 * and other support functionality.
 * 
 * Thanks are in order for the excellent serial tutorial at:
 *     https://www.cmrr.umn.edu/~strupp/serial.html
 */

#include "unix.h"


// Globals
int      verbose = 0;           // Flag: Enable verbose output
int      quiet   = 0;           // Flag: Disable all output

uint16_t timeout = 1000;        // Port read timeout, in ms


int main (int argc, char *argv[]) {
    
    int      retval;            // Function return value
    int      fd;                // Port file descriptor
    
    // Flags
    uint8_t  flags = 0;         // Flags bitmap
    uint8_t  mode  = 0;         // Mode bitmap
    
    // Device information
    uint32_t partid;            // Part ID
    uint32_t uid[4];            // Unique ID
    uint8_t  romver[2];         // ROM version
    uint32_t baud;              // Baud rate as a number
    int      baudconst;         // Baud rate as a constant
    
    // CLI processing stuff
    int      opt;               // Current option
    int      idx = 0;           // Current option index
    
    // CLI options for getopt_long()
    struct option long_opts[] = {
        // Serial options
        { "port",        required_argument, NULL,     'p' },
        { "timeout",     required_argument, NULL,     't' },
        { "baud",        required_argument, NULL,     'b' },
        { "stop-bits",   required_argument, NULL,     's' },
        // Device options
        { "sync",        required_argument, NULL,     'a' },
        { "clock",       required_argument, NULL,     'c' },
        { "part-id",     required_argument, NULL,     'i' },
        { "unique-id",   required_argument, NULL,     'u' },
        { "rom-version", required_argument, NULL,     'r' },
        { "force",       no_argument,       NULL,      0  },
        { "unlock",      required_argument, NULL,     'l' },
        { "no-erase",    no_argument,       NULL,      0  },
        // Output options
        { "verbose",     no_argument,       &verbose,  1  },
        { "quiet",       no_argument,       &quiet,    1  },
        { "help",        no_argument,       NULL,      0  },
        { "version",     no_argument,       NULL,      0  },
        // Query operations
        { "get-pid",     no_argument,       NULL,      0  },
        { "get-uid",     no_argument,       NULL,      0  },
        { "get-version", no_argument,       NULL,      0  },
        // Flash operations
        { "check",       no_argument,       NULL,      0  },
        { "erase",       no_argument,       NULL,      0  },
        { "read",        no_argument,       NULL,      0  },
        { "write",       no_argument,       NULL,      0  },
        { "compare",     no_argument,       NULL,      0  },
        // List terminator
        { NULL,          0,                 NULL,      0  }
    };
    
    // Configuration as requested by CLI
    struct cfgopt_t config = {
        .port     = NULL,
        .baud     = 9600,
        .stop     = 1,
        .attempts = 30,
        .clock    = 0,
    };
    
    
    /* - Check command line parameters ------------------------------------ */
    
    // Loop through each parameter
    while (1) {
        
        opt = getopt_long(argc, argv, "p:t:b:s:a:c:i:u:r:fl:e", long_opts, &idx);
        
        // No more options?
        if (opt == -1) break;
        
        switch (opt) {
        
        case 'p':       // Serial port
            config.port = optarg;
            break;
            
        case 't':       // Read timeout
            sscanf(optarg, "%"SCNu16, &timeout);
            break;
            
        case 'b':       // Baud rate
            sscanf(optarg, "%"SCNu32, &config.baud);
            break;
            
        case 's':       // Stop bits
            sscanf(optarg, "%"SCNu8, &config.stop);
            break;
            
        case 'a':       // Sync attempts
            sscanf(optarg, "%"SCNu8, &config.attempts);
            break;
            
        case 'c':       // MCU Clock freq
            sscanf(optarg, "%"SCNu32, &config.clock);
            break;
            
        case 'i':       // Part ID
            sscanf(optarg, "%x", &config.pid);
            break;
            
        case 'u':       // Unique ID
            sscanf( optarg, "%x,%x,%x,%x",
                    &config.uid[0], &config.uid[1],
                    &config.uid[2], &config.uid[3] );
            break;
            
        case 'r':       // ROM version
            sscanf( optarg, "%"SCNu8".%"SCNu8,
                    &config.romver[0], &config.romver[1] );
            break;
            
        case 'f':       // Force (continue w/o ID check)
            flags |= FLAG_FORCE;
            break;
            
        case 'l':       // Unlock code
            sscanf(optarg, "%"SCNu16, &config.lock);
            break;
            
        case 'e':       // Don't erase flash before write
            flags &= ~FLAG_ERASE;
            break;
            
        case 0:         // Long option
            
            //printf("Found flag %i, '%s'\n", idx, long_opts[idx].name);
            
            // Program options
            if (strcmp(long_opts[idx].name, "help") == 0) {
                flags |= FLAG_HELP;
                break;
            }
            
            if (strcmp(long_opts[idx].name, "version") == 0) {
                flags |= FLAG_VERSION;
                break;
            }
            
            
            // Query operations
            if (strcmp(long_opts[idx].name, "get-pid") == 0) {
                mode |= (MODE_QUERY | MODE_CHECK_PID);
                break;
            }
            
            if (strcmp(long_opts[idx].name, "get-uid") == 0) {
                mode |= (MODE_QUERY | MODE_CHECK_UID);
                break;
            }
            
            if (strcmp(long_opts[idx].name, "get-romver") == 0) {
                mode |= (MODE_QUERY | MODE_CHECK_ROM);
                break;
            }
            
            
            // Flash operations
            if (strcmp(long_opts[idx].name, "check") == 0) {
                mode |= MODE_FLASH;
                break;
            }
            
            if (strcmp(long_opts[idx].name, "erase") == 0) {
                mode |= (MODE_FLASH | MODE_FLASH_WRITE);
                break;
            }
            
            if (strcmp(long_opts[idx].name, "read") == 0) {
                mode |= MODE_FLASH;
                break;
            }
            
            if (strcmp(long_opts[idx].name, "write") == 0) {
                mode |= (MODE_FLASH | MODE_FLASH_WRITE);
                break;
            }
            
            if (strcmp(long_opts[idx].name, "compare") == 0) {
                mode |= MODE_FLASH;
                break;
            }
            
            break;
            
        default:        // ??
            // Note: getopt_long() will output an error for us.
            // This can be disabled by setting opterr = 0 before the loop.
            
            //fprintf(stderr, "Unrecognized option: '%c', %s\n", optopt, optarg);
            return 1;
        }
    }
    
    
    // Check for syntax help request
    if (flags & FLAG_HELP) {
        show_usage(argv[0]);
        show_help();
        
        return 0;
    }
    
    // Check for version info request
    if (flags & FLAG_VERSION)  {
        show_version();
        
        return 0;
    }
    
    
    /* - Validate configuration ------------------------------------------- */
    
    // Demand required parameters
    if (config.port == NULL) {
        show_usage(argv[0]);
        return 1;
    }
    
    if (config.clock == 0) {
        show_usage(argv[0]);
        return 1;
    }
    
    
    // Validate optional parameters
    if ( (timeout < 100) || (timeout > 10000) ) {
        fprintf( stderr, "Timeout (-t) value is invalid: %"PRIu32"\n",
                 timeout );
        return 1;
    }
    
    if ( (config.stop != 1) && (config.stop != 2) ) {
        fprintf( stderr, "Stop bits (-s) value is invalid: %"PRIu8"\n",
                 config.stop );
        return 1;
    }
    
    
    /* - Sort out operational modes --------------------------------------- */
    
    // Check for operation sanity
    if ( (mode & MODE_FLASH) && (mode & MODE_QUERY) ) {
        fprintf( stderr,
                 "Query and flash operations cannot be used together\n" );
        return 1;
    }
    
    // Set ID check flags based on which IDs were given
    // (This isn't done during parsing to avoid conflicts when --part-id
    // and co are specified along with --get-pid and friends)
    if (!(mode & MODE_QUERY)) {
        
        if (config.pid)
            mode |= MODE_CHECK_PID;
        
        if (config.uid[0] | config.uid[1] | config.uid[2] | config.uid[3])
            mode |= MODE_CHECK_UID;
        
        if (config.romver[0] | config.romver[1])
            mode |= MODE_CHECK_ROM;
    }
    
    // Make sure write operations are validated with an ID check
    if ( (mode & MODE_FLASH_WRITE) && !(flags & FLAG_FORCE)
         && ( !(mode & (MODE_CHECK_PID | MODE_CHECK_UID)) ) )
    {
        fprintf( stderr,
                 "Flash operations require either --part-id or --unique-id "
                 "for verification.\n" );
        return 1;
    }
    
    if ( (mode & MODE_FLASH_WRITE) && (config.lock = 0x00) ) {
        fprintf( stderr, "Flash operations require an --unlock code\n");
        return 1;
    }
    
    
    // TESTING:
    if (verbose) printf("flags=0x%02x, mode=0x%02x\n", flags, mode);
    
    
    /* - Open and configure serial interface ------------------------------ */
    
    // Open the serial port
    fd = open(config.port, O_RDWR | O_NOCTTY | O_NDELAY);
    
    if (fd == -1) {
        
        if (!quiet) {
            fprintf( stderr, "Couldn't open port '%s': (%i) %s\n",
                     config.port, errno, strerror(errno) );
        }
        
        return 1;
    }
    
    
    if (verbose) printf("Serial port %s opened\n", config.port);
    
    
    // Configure for 9600,8N1, XON/XOFF, raw mode
    baud      = 9600;
    baudconst = get_baud(&baud);
    
    set_port_config(fd, baudconst, 1);
    
    
    /* - Start session with device boot ROM ------------------------------- */
    
    // Attempt to synchronize
    if (!quiet) {
        printf("Synchronizing... ");
        fflush(stdout);
    }
    
    while (1) {
        
        if (!quiet) spinner(1);
        
        retval = synchronize(fd);
        
        // Success -- clear progress indicator and leave loop
        if (retval > 0) {
            if (!quiet) spinner(0);
            break;
        }
        
        // Write error or somesuch
        if (retval < 0) {
            
            if (!quiet) {
                spinner(0);
                fprintf(stderr, "Failed: (%i) %s\n", errno, strerror(errno));
            }
            
            return 1;
        }
        
        // Still here?  retval == 0, i.e., timed out
        
        // Requested infinite attempts
        if (config.attempts == 0) continue;
        
        // Limited attempts -- continue if credits remain
        config.attempts--;
        if (config.attempts) continue;
        
        // Out of luck :-(
        if (!quiet) {
            spinner(0);
            fprintf(stderr, "Failed: (%i) %s\n", errno, strerror(errno));
        }
        
        return 1;
    }
    
    if (!quiet) {
        printf("OK\nSetting clock frequency... ");
        fflush(stdout);
    }
    
    if (!set_clock(fd, config.clock)) {
        
        if (!quiet) {
            fprintf( stderr, "Failed: (%i) %s\n", errno, strerror(errno) );
        }
        
        return 1;
    }
    
    if (!quiet) {
        printf("OK\nDisabling echo... ");
        fflush(stdout);
    }
    
    if (!echo_off(fd)) {
        
        if (!quiet) {
            fprintf( stderr, "Failed: (%i) %s\n", errno, strerror(errno) );
        }
        
        return 1;
    }
    
    if (!quiet) printf("OK\n");
    
    
    /* - Configure device ------------------------------------------------- */
    
    if ( (config.baud != 9600) || (config.stop != 1) ) {
        
        // Get supported baud rate
        baud      = config.baud;
        baudconst = get_baud(&baud);
        
        if (!quiet) {
            
            if (baud != config.baud) {
                
                fprintf( stderr,
                         "Warning! Requested baud rate (%"PRIu32") is "
                         "unavailable.  Using %"PRIu32" instead\n",
                         config.baud, baud );
            }
            
            printf( "Setting new port configuration "
                    "(%"PRIu32" baud, %u stop bits)... ",
                    baud, config.stop );
            
            fflush(stdout);
        }
        
        if (!set_new_speed(fd, baud, config.stop)) {
            
            if (!quiet) {
                fprintf( stderr, "Failed: (%i) %s\n",
                         errno, strerror(errno) );
            }
            
            return 1;
        }
        
        // Set local port speed to match
        set_port_config(fd, baudconst, config.stop);
        
        if (!quiet) printf( "OK\n");
    }
    
    
    /* - Query for device info--------------------------------------------- */
    
    if (!quiet) printf("Getting part ID... ");
    
    if (!get_part_id(fd, &partid)) {
        if (!quiet)
            fprintf( stderr, "Failed: (%i) %s\n", errno, strerror(errno) );
        
        return 1;
    }
    
    if ( !quiet || ((mode & MODE_QUERY) && (mode & MODE_CHECK_PID)) )
        printf("0x%8x\n", partid);
    
    if ((mode & MODE_QUERY) && (mode & MODE_CHECK_PID))
        return 0;
    
    
    if (!quiet) printf("Getting unique ID... ");
    
    if (!get_unique_id(fd, uid)) {
        if (!quiet)
            fprintf( stderr, "Failed: (%i) %s\n", errno, strerror(errno) );
        
        return 1;
    }
    
    if (!quiet || ((mode & MODE_QUERY) && (mode & MODE_CHECK_UID)) )
        printf( "0x%08x 0x%08x 0x%08x 0x%08x\n", uid[0], uid[1], uid[2], uid[3]);
    
    if ((mode & MODE_QUERY) && (mode & MODE_CHECK_UID))
        return 0;
    
    
    if (!quiet) printf("Getting ROM version... ");
    
    if (!get_rom_version(fd, romver)) {
        if (!quiet)
            fprintf( stderr, "Failed: (%i) %s\n", errno, strerror(errno) );
        
        return 1;
    }
    
    if ( !quiet || ((mode & MODE_QUERY) && (mode & MODE_CHECK_ROM)) )
        printf("%1u.%1u\n", romver[0], romver[1]);
    
    if ((mode & MODE_QUERY) && (mode & MODE_CHECK_ROM))
        return 0;
    
    
    /* - Perform requested flash operations ------------------------------- */
    
    // Validate ID/version
    if (mode & MODE_FLASH_WRITE) {
        
        if ( (mode & MODE_CHECK_PID) && (config.pid != partid) ) {
            if (!quiet) {
                fprintf( stderr,
                         "Part ID does not match! "
                         "(Expected 0x%08x, got 0x%08x)\n",
                         config.pid, partid );
            }
            
            return 1;
        }
        
        if ( (mode & MODE_CHECK_UID) &&
             (config.uid[0] != uid[0]) && (config.uid[1] != uid[1]) &&
             (config.uid[2] != uid[2]) && (config.uid[3] != uid[3]) )
        {
            if (!quiet) {
                fprintf( stderr,
                         "Unique ID does not match! "
                         "(Expected 0x%08x 0x%08x 0x%08x 0x%08x, "
                         "got 0x%08x 0x%08x 0x%08x 0x%08x)\n",
                         config.uid[0], config.uid[1],
                         config.uid[2], config.uid[3],
                         uid[0], uid[1], uid[2], uid[3] );
            }
            
            return 1;
        }
        
        if ( (mode & MODE_CHECK_ROM) &&
             (config.romver[0] != romver[0]) &&
             (config.romver[1] != romver[1]) )
        {
            if (!quiet) {
                fprintf( stderr,
                         "ROM version does not match! (Expected "
                         "%"PRIu8".%"PRIu8", got %"PRIu8".%"PRIu8")\n",
                         config.romver[0], config.romver[1],
                         romver[0], romver[1] );
            }
            
            return 1;
        }
    }
    
    
    if (mode & MODE_FLASH_WRITE) {
        
        // Unlock flash
        if (!quiet) {
            printf("Unlocking flash with code %"PRIu32"... ", config.lock);
            fflush(stdout);
        }
        
        if (!unlock(fd, config.lock)) {
            
            if (!quiet) {
                fprintf( stderr, "Failed: (%i) %s\n", errno, strerror(errno) );
            }
            
            return 1;
        }
        
        if (!quiet) printf("OK\n");
    }
    
    
    /* - Clean up --------------------------------------------------------- */
    
    // Close the serial port
    close(fd);
    
    if (verbose) printf("Port closed\n");
    
    return 0;
}


void show_usage (char *prog) {
    printf("Usage: %s [options] -p <port> -c <clock> [operations]\n", prog);
}


void show_help () {
    
    printf(
"\nSerial Options:\n"
"  -p  --port        Serial port (e.g., /dev/ttyXX)\n"
"  -t  --timeout     Read timeout (in ms, 100-10000, default 1000)\n"
"  -b  --baud        Set serial port speed (default 9600)\n"
"  -s  --stop-bits   Set serial stop bits (default 1)\n"
"\nDevice Options:\n"
"  -a  --sync        Sync attempts (# of reads, default 30, 0=infinite)\n"
"  -c  --clock       MCU clock frequency (in kHz)\n"
"  -i  --part-id     Verify part ID\n"
"  -u  --unique-id   Verify part unique ID\n"
"  -r  --rom-version Verify ROM version\n"
"  -f  --force       Permit flash operations without checking ID\n"
"  -l  --unlock      Unlock code (req'd to modify flash)\n"
"  -e  --no-erase    Don't erase flash before writing\n"
"\nOutput Options:\n"
"      --verbose     Show more progress info\n"
"      --quiet       Don't show any progress info\n"
"      --help        Show usage info\n"
"      --version     Report program version and exit\n"
"\nQuery Operations:\n"
"      --get-pid     Get part ID and exit\n"
"      --get-uid     Get unique ID and exit\n"
"      --get-version Get ROM version and exit\n"
"\nFlash Operations:\n"
"      --check       Check for blank flash sectors\n"
"      --erase       Erase flash sectors\n"
"      --read        Read data from memory/flash to file\n"
"      --write       Write data from file to memory/flash\n"
"      --compare     Compare data between file and memory/flash\n"
"\nPort Speed:\n"
"Session init always takes place at 9600 baud, 8 data bits, 1 stop bit, and no\n"
"parity.  If --baud and/or --stop-bits are specified, the changes will apply\n"
"shortly after initialization.\n"
"\nID and Version Matching:\n"
"Either the part ID or unique ID must be verified before performing any\n"
"operations that read from or write to memory.  This check can be bypassed by\n"
"passing the -f option, but there should rarely be a valid need to do so.\n"
"\nQuery Operations:\n"
"Query operations are intended to be used with the --quiet switch for easy\n"
"processing in scripts.  Only one query can be used at a time.\n"
"\nFlash Operations:\n"
"Flash operations will be performed in the order listed: check, erase, read,\n"
"write, compare.  Processing will stop when any condition returns unsuccess-\n"
"fully (including check returning non-empty sectors.)\n"
"Flash sectors selected for a write operation will be automatically erased\n"
"unless the --no-erase option is specified.  As such, there is no need to\n"
"manually --erase sectors before writing.\n"
"\n"
);
}


// Show program version info
void show_version () {
    printf("serial test app v%01u.%01u, rev.%01u\n", VMAJOR, VMINOR, VREV);
}


// Draw a dynamic progress indicator
void spinner (uint8_t spin) {
    
    static char next = 0x00;
    
    // Clear and reset
    if (!spin) {
        next = 0x00;
        printf("\b ");
        fflush(stdout);
        return;
    }
    
    switch (next) {
    
    case 0x00:              // Initial
        next = '|';
        printf("|");
        fflush(stdout);
        return;
        
    case '|':
        next = '/';
        break;
        
    case '/':
        next = '-';
        break;
        
    case '-':
        next = '\\';
        break;
        
    default:
        next = '|';
        break;
    }
    
    printf("\b%c", next);
    fflush(stdout);
}    


// Get nearest baud rate constant from specified rate
int get_baud (uint32_t *baud) {
    
    // Maps actual baudrates to constants
    static struct baudpair_t bauds[] = {
        { 50,     B50 },
        { 75,     B75 },
        { 110,    B110 },
        { 134,    B134 },
        { 150,    B150 },
        { 200,    B200 },
        { 300,    B300 },
        { 600,    B600 },
        { 1200,   B1200 },
        { 1800,   B1800 },
        { 2400,   B2400 },
        { 4800,   B4800 },
        { 9600,   B9600 },
        { 19200,  B19200 },
        { 38400,  B38400 },
        { 57600,  B57600 },
        { 115200, B115200 }
    };
    
    uint8_t  max_idx = (sizeof(bauds) / sizeof(struct baudpair_t));
    uint8_t  last_idx = 0;
    uint32_t last_error;
    uint32_t error;
    uint8_t  i;
    
    // Find nearest baud rate constant
    for (i = 0; i < max_idx; i++) {
        
        // Perfect match?  We're done here
        if (bauds[i].baud == *baud)
            return bauds[i].constant;
        
        // Not so perfect match, keep track of error
        if (*baud >= bauds[i].baud) {
            last_idx   = i;
            last_error = (*baud - bauds[i].baud);
        }
    }
    
    // Cap at highest possible speed
    if (last_idx == (max_idx - 1)) {
        *baud = bauds[last_idx].baud;
        return bauds[last_idx].constant;
    }
    
    // Which is closer -- below or above?
    error = bauds[last_idx + 1].baud - *baud;
    
    if (error <= last_error) last_idx++;
    
    *baud = bauds[last_idx].baud;
    
    return bauds[last_idx].constant;
}


// Configure local serial interface
void set_port_config (int fd, int baudconst, uint8_t stop) {
    
    struct termios config;
    
    // Get current port configuration
    tcgetattr(fd, &config);
    
    // Set baud rate
    cfsetispeed(&config, baudconst);
    cfsetospeed(&config, baudconst);
    
    // Set ownership, character size, stop bit, parity...
    config.c_cflag &= ~(CSIZE | CSTOPB | PARENB);
    config.c_cflag |= (CLOCAL | CREAD | CS8);
    
    if (stop == 2) config.c_cflag |= CSTOPB;
    
    // Disable job signals and echo, enable raw mode
    config.c_lflag &= ~(ISIG | ECHO | ECHOE | ICANON);
    
    // Enable software flow control
    config.c_iflag &= ~IXANY;
    config.c_iflag |= (IXON | IXOFF);
    
    config.c_cc[VSTOP]  = 0x13; // DC3
    config.c_cc[VSTART] = 0x11; // DC1
    
    // Don't process output
    config.c_oflag &= ~(OPOST | ONLCR);
    
    // Set port configuration
    tcsetattr(fd, TCSANOW, &config);
}