#include "uuencode.h"

// Encodes a line of characters from up to 94 bytes of input data
void uu_encode_line (const char *in, char *out, uint8_t datalen, uint32_t *cksum) {
    
    char    tmp[3];     // Data to be encoded
    uint8_t len;        // Bytes remaining
    uint8_t pos;        // Buffer offset
    uint8_t i;          // Loop counter
    
    
    // Don't exceed configured line length
    if (datalen > UU_DATAMAX) {     // BTW, >94 would exceed 7-bit barrier
        *out = 0x00;
        return;
    }
    
    
    // Start output line with the input data length
    *out = (char)(datalen + 0x20); // offset ensures a printable char
    
    
    // Encode data in blocks of three bytes
    len = datalen;
    pos = 0;
    
    while (len) {
        
        // Encode next block of three bytes
        if (len > 3) {
            
            uu_encode(in + pos, out + ((pos / 3) * 4) + 1);
            
            len -=3;
            pos +=3;
            
            continue;
        }
        
        // Incomplete blocks must be padded with nulls
        for (i = 0; i < 3; i++)
            tmp[i] = (len > i) ? (char)*(in + pos + i) : 0x00;
        
        // Encode this block of three bytes
        uu_encode(tmp, out + ((pos / 3) * 4) + 1);
        
        len = 0;
        pos += 3;
    }
    
    // Terminate line with a null
    *(out + ((pos / 3) * 4) + 1) = 0x00;
    
    
    // Update checksum, if requested
    if (cksum != NULL)
        *cksum += uu_checksum(in, datalen);
}


// Decodes a line of characters back into binary data
uint8_t uu_decode_line (const char *in, char *out, uint32_t *cksum) {
    
    uint8_t len;        // Blocks remaining
    uint8_t datalen;    // Bytes expected
    uint8_t pos;        // Buffer offset
    uint8_t i;          // Loop counter
    
    
    // Read length byte
    datalen = ((uint8_t)*in) - 0x20;
    
    
    // Stop if line would exceed configured length
    if (datalen > UU_DATAMAX)
        return 0;
    
    
    // Decode data
    pos = 0;
    len = (datalen / 3)             // Datalen refers to 8-bit input bytes
        + ((datalen % 3) ? 1 : 0);  // ...account for padding bytes
    
    while (len) {
        
        // Check for premature end-of-data
        for (i = 0; i < 4; i++) {
            if (*(in + pos + i) == 0x00) break;
        }
        
        // Decode block
        uu_decode(in + pos + 1, out + ((pos / 4) * 3));
        
        len--;
        pos += 4;
    }
    
    
    // Trim nulls if necessary
    if (datalen < pos) pos = datalen;
    
    // Update checksum, if requested
    if (cksum != NULL)
        *cksum += uu_checksum(out, pos);
    
    // Return number of data bytes decoded
    return pos;
}


// Encode 3x 8-bit input bytes to 4x 6-bit output chars
void uu_encode (const char *in, char *out) {
    
    uint8_t i;
    
    //  |012345 67|8901 2345|67 890123| <-- 8-bit input bytes
    //  |012345|67 8901|2345 67|890123| <-- 6-bit output chars
        
    // Divide 24-bit input into 6-bit groups
    *(out + 0) = (char)((*(in + 0) & 0xfc) >> 2);   // Bits 0-5
    *(out + 1) = (char)((*(in + 0) & 0x03) << 4)    // Bits 6-7
               | (char)((*(in + 1) & 0xf0) >> 4);   // Bits 8-11
    *(out + 2) = (char)((*(in + 1) & 0x0f) << 2)    // Bits 12-15
               | (char)((*(in + 2) & 0xc0) >> 6);   // Bits 16-17
    *(out + 3) = (char) (*(in + 2) & 0x3f);         // Bits 18-23
    
    // Offset values by 0x60 if char == 0x00, or 0x20 otherwise
    for (i = 0; i < 4; i++)
        *(out + i) += (*(out + i) == 0x00) ? 0x60 : 0x20;
}


// Decode 4x 6-bit input chars to 3x 8-bit output bytes
void uu_decode (const char *in, char *out) {
    
    uint8_t i;
    char    tmp[4];
    
    // Move values back down to their original values
    for (i = 0; i < 4; i++)
        tmp[i] = (*(in + i)) - ((*(in + i) == 0x60) ? 0x60 : 0x20);
    
    // Recombine 6-bit groups into 24-bit output stream
    *(out + 0) = (char)((tmp[0] & 0x3f) << 2)    // Bits 0-5
               | (char)((tmp[1] & 0x30) >> 4);   // Bits 6-7
    *(out + 1) = (char)((tmp[1] & 0x0f) << 4)    // Bits 8-11
               | (char)((tmp[2] & 0x3c) >> 2);   // Bits 12-15
    *(out + 2) = (char)((tmp[2] & 0x03) << 6)    // Bits 16-17
               | (char) (tmp[3] & 0x3f);         // Bits 18-23
}


// Generate checksum from 8-bit input bytes
uint32_t uu_checksum (const char *in, uint16_t len) {
    
    uint32_t sum = 0;
    
    // Loop through input bytes, summing their values
    while (len) {
        sum += (uint8_t)*in;
        len--;
        in++;
    }
    
    return sum;
}